import hashproduct.popmatch.*;
import static hashproduct.popmatch.Statics.*;
import static hashproduct.popmatch.threesat.Statics.*;

import java.util.*;
import static java.lang.Math.*;

import hashproduct.popmatch.threesat.ReductionToPopMatch;

// Testing various things leaves lots of unused code.
// Don't bother me about it.
@SuppressWarnings("unused")
public class PopularMatcherMain {
	private static final int I = IMPOSSIBLE_RANK;
	
	/*static Instance i;
	static Matching m, m2;
	static MatchingAnalysis ma;
	static Classification c1, c2;
	static Matching c1m, c2m;*/
	
	private static void latexPrintInstance(Instance i) {
		System.out.println("\\pmatrix{");
		for (int person = 0; person < i.numPeople(); person++) {
			for (int post = 0; post < i.numPosts(); post++) {
				System.out.print((i.rankFor(person, post) == IMPOSSIBLE_RANK ? "-" : i.rankFor(person, post) + 1));
				if (post == i.numPosts() - 1)
					System.out.println("\\cr");
				else
					System.out.print("&");
			}
		}
		System.out.println("}");
	}
	
	private static int investigateInstance(Instance i) {
		StringBuilder sb = new StringBuilder();
		sb.append("Instance: " + Statics.toString(i) + "\n");
		Solver[] solvers = Solvers.factorSolvers;
		PressureAnalysis[] mas = new PressureAnalysis[solvers.length];
		
		int bestUf = Integer.MAX_VALUE;
		for (int s = 0; s < solvers.length; s++) {
			mas[s] = Pressurizer.pressurize(solvers[s].solve(i));
			bestUf = min(bestUf, mas[s].unpopularity);
		}
		boolean loses = false, wins = false;
		for (int s = 0; s < solvers.length; s++) {
			if (mas[s].unpopularity != bestUf)
				if (solvers[s] == ClassMaximalMatcher.it)
					loses = true;
				else
					wins = true;
			sb.append("-- " + solvers[s].solverName()
				+ (mas[s].unpopularity == bestUf ? "" : " (SUBOPTIMAL)") + "\n");
			sb.append(mas[s].toString());
		}
		
		if (loses)
			System.out.println("*** CMM LOSES ***");
		//if (loses || wins)
			System.out.println(sb.toString());
		return bestUf;
	}
	
	private static void investigateRandomInstances(int howMany, int numPeople, int numPosts, int numRanks, Random r) {
		for (int test = 0; test < howMany; test++) {
			Instance i = Instances.randomInstance(numPeople, numPosts, numRanks, r);
			investigateInstance(i);
		}
	}
	
	private static void marginInvestigateRandomInstances(int howMany, int numPeople, int numPosts, int numRanks, Random r) {
		for (int test = 0; test < howMany; test++) {
			System.out.println("Instance " + test);
			Instance i = Instances.randomInstance(numPeople, numPosts, numRanks, r);
			marginInvestigateInstance(i);
		}
	}
	
	private static void testBox(int[][] builtinRanks, int[] wideRanks, int[] narrowRanks) {
		int numBuiltins = builtinRanks.length;
		int u1, u2, u3, u4, u5, u6;
		
		{
		System.out.println("== Box case none");
		Instance i1 = new InstanceImpl(builtinRanks);
		u1 = investigateInstance(i1);
		}
		
		{
		System.out.println("== Box case W");
		int[][] ranks2 = new int[numBuiltins + 1][];
		System.arraycopy(builtinRanks, 0, ranks2, 0, numBuiltins);
		ranks2[numBuiltins] = wideRanks;
		Instance i2 = new InstanceImpl(ranks2);
		u2 = investigateInstance(i2);
		}
		
		{
		System.out.println("== Box case N");
		int[][] ranks3 = new int[numBuiltins + 1][];
		System.arraycopy(builtinRanks, 0, ranks3, 0, numBuiltins);
		ranks3[numBuiltins] = narrowRanks;
		Instance i3 = new InstanceImpl(ranks3);
		u3 = investigateInstance(i3);
		}
		
		{
		System.out.println("== Box case WN");
		int[][] ranks4 = new int[numBuiltins + 2][];
		System.arraycopy(builtinRanks, 0, ranks4, 0, numBuiltins);
		ranks4[numBuiltins] = wideRanks;
		ranks4[numBuiltins + 1] = narrowRanks;
		Instance i4 = new InstanceImpl(ranks4);
		u4 = investigateInstance(i4);
		}
		
		{
		System.out.println("== Box case NN");
		int[][] ranks5 = new int[numBuiltins + 2][];
		System.arraycopy(builtinRanks, 0, ranks5, 0, numBuiltins);
		ranks5[numBuiltins] = narrowRanks;
		ranks5[numBuiltins + 1] = narrowRanks;
		Instance i5 = new InstanceImpl(ranks5);
		u5 = investigateInstance(i5);
		}
		
		{
		System.out.println("== Box case WNN");
		int[][] ranks6 = new int[numBuiltins + 3][];
		System.arraycopy(builtinRanks, 0, ranks6, 0, numBuiltins);
		ranks6[numBuiltins] = wideRanks;
		ranks6[numBuiltins + 1] = narrowRanks;
		ranks6[numBuiltins + 2] = narrowRanks;
		Instance i6 = new InstanceImpl(ranks6);
		u6 = investigateInstance(i6);
		}
		
		System.out.println("== SUMMARY: _ W N WN NN WNN");
		System.out.println("Actual:     " + u1 + " " + u2 + " " + u3 + " " + u4 + "  " + u5 + "  " + u6);
		System.out.println("Desired:    2 2 2 3  2  3");
	}
	
	private static void breakIBHP() {
		int ibhpLuf = 5;
		Instance i = Instances.ibhpFoo(ibhpLuf);
		investigateInstance(i);
		Matching m3 = Instances.ibhpFooMatch(ibhpLuf);
		PressureAnalysis ma3 = Pressurizer.pressurize(m3);
		System.out.println(ma3.toString());
	}
	
	private static void testTheBox() {
		// Box used in the 3SAT reduction
		testBox(
			new int[][] {
				{1, 2, 3, 4},
				{1, 2, 3, 4},
				{1, 2, 3, 4},
			},
			new int[] {1, 2, 4, 3},
			new int[] {I, I, I, 1}
			);
		
		// An attempt to simplify the box so it works between pressures 1 and 2,
		// but a pressure on a linking post breaks it, so we can't show P = NP
		testBox(
			new int[][] {
				{1, 2, 3},
				{1, 2, 3},
			},
			new int[] {1, 3, 2},
			new int[] {I, I, 1}
			);
	}
	
	private static hashproduct.popmatch.threesat.Instance
		randomThreesat(int vars, int stmts, Random rnd1) {
		int[][] refs = new int[stmts][3];
		for (int s = 0; s < stmts; s++) {
			for (int p = 0; p < 3; p++) {
				int r1 = rnd1.nextInt(2 * vars);
				if (r1 >= vars)
					r1 = ~(r1 - vars);
				refs[s][p] = r1;
			}
		}
		return new hashproduct.popmatch.threesat.InstanceImpl(vars, refs);
	}
	
	// Shows results of both factor and margin analyses.
	// Code stolen from their respective #toString methods.
	private static String hybridFormat(Matching m, PressureAnalysis pa, MarginAnalysis ma) {
		StringBuilder sb = new StringBuilder();
		sb.append("Unpopularity factor: ");
		sb.append(pa.unpopularity == PARETO_INEFFICIENT
			? "Pareto inefficient" : pa.unpopularity);
		sb.append('\n');
		sb.append("Unpopularity margin: ");
		sb.append(ma.unpopularity);
		sb.append('\n');
		sb.append("Numbers give pressures; <>s give margin's better matching.\n");
		Statics.format(m.instance(), m, ma.better, pa.c, sb);
		return sb.toString();
	}
	private static String hybridAnalyzeAndFormat(Matching m) {
		PressureAnalysis pa = Pressurizer.pressurize(m);
		MarginAnalysis ma = Marginizer.marginize(m);
		return hybridFormat(m, pa, ma);
	}
	
	private static void testReductionOnTVs(ReductionToPopMatch reduction,
		boolean satisfiable, boolean[] truthVals) {
		hashproduct.popmatch.threesat.Instance tsi = reduction.tsi();
		boolean satisfied = satisfies(truthVals, tsi);
		Matching m = reduction.truthValuesToMatching(truthVals);

		PressureAnalysis pa = Pressurizer.pressurize(m);
		boolean factorGood = (pa.unpopularity == reduction.goodUnpopFactor());
		boolean factorPass = (pa.unpopularity >= reduction.goodUnpopFactor() && factorGood == satisfied);
		MarginAnalysis ma = Marginizer.marginize(m);
		boolean marginGood = (ma.unpopularity == reduction.goodUnpopMargin());
		boolean marginPass = (ma.unpopularity >= reduction.goodUnpopMargin() && marginGood == satisfied);
		
		if (factorPass && marginPass) {
			System.out.println("Truth values " + Arrays.toString(truthVals) + ": pass");
		} else {
			System.err.println("Truth values " + Arrays.toString(truthVals) + ": FAIL");
			System.err.println(hybridFormat(m, pa, ma));
		}
	}
	
	private static void testReduction1Rec(ReductionToPopMatch reduction,
		boolean satisfiable, boolean[] truthVals, int varStart) {
		if (varStart == reduction.tsi().numVariables())
			testReductionOnTVs(reduction, satisfiable, truthVals);
		else {
			truthVals[varStart] = false;
			testReduction1Rec(reduction, satisfiable, truthVals, varStart+1);
			truthVals[varStart] = true;
			testReduction1Rec(reduction, satisfiable, truthVals, varStart+1);
		}
	}
	
	private static void testReduction1(
		hashproduct.popmatch.threesat.Instance tsi) {
		ReductionToPopMatch reduction = new ReductionToPopMatch(tsi);
		System.out.println("Testing reduction on 3SAT instance: " + tsi.toString());
		boolean[] soln = solution(tsi);
		if (soln == null)
			System.out.println("Not satisfiable");
		else
			System.out.println("Solution: " + Arrays.toString(soln));
		boolean satisfiable = (soln != null);
		boolean[] truthVals = new boolean[tsi.numVariables()];
		testReduction1Rec(reduction, satisfiable, truthVals, 0);
	}
	
	private static void testReduction() {
		for (int i = 0; i < 20; i++) {
			System.out.println("3SAT INSTANCE #" + i);
			hashproduct.popmatch.threesat.Instance tsi =
				randomThreesat(3, 7, rnd);
			testReduction1(tsi);
		}
	}
	
	private static void threesatDemo() {
		hashproduct.popmatch.threesat.Instance tsi =
			new hashproduct.popmatch.threesat.InstanceImpl(2, 
				new int[][] {
				{0, 1, 0},
				{~1, ~0, ~1},
				});
		ReductionToPopMatch trans = new ReductionToPopMatch(tsi);
		System.out.println(trans.gpmi());
		
		// Fails to satisfy the second statement
		Matching m1 = trans.truthValuesToMatching(new boolean[] {true, true});
		System.out.println(hybridAnalyzeAndFormat(m1));
		
		// Satisfies both statements
		Matching m2 = trans.truthValuesToMatching(new boolean[] {true, false});
		System.out.println(hybridAnalyzeAndFormat(m2));
		
		//investigateInstance(trans.gpmi());
		//latexPrintInstance(trans.gpmi());
	}
	
	private static final Random rnd = new Random(314159);
	
	private static void oldMainStuff() {
		//investigateRandomInstances(5000, 4, 4, 4, rnd);
		
		// I don't remember what was so special about these instances.
		/*Instance i = new InstanceImpl(5, 5,
			new int[][] {{0, 1, 2, 3, 1}, {0, 1, 2, 3, I}, {0, 1, 2, 3, I}, {0, 1, 2, 3, I}, {I, I, I, I, 1}});*/
		/*Instance i = new InstanceImpl(4, 4, new int[][] {{2, 2, I, 0}, {1, I, 3, 3}, {2, 0, 1, 1}, {3, 2, I, 0}});*/
		//Instance i = new InstanceImpl(3, 3, new int[][] {{2, 1, 0}, {2, 1, 0}, {I, 2, 1}});
		//Instance i = new InstanceImpl(3, 3, new int[][] {{2, 1, 0}, {1, 2, 1}, {2, I, 0}});
		
		/*
		Instance i = new InstanceImpl(new int[][] {
			{1, 0, I, 2},
			{1, 0, I, 2},
			{I, 0, 1, 2},
			{I, 0, 1, 2},
		});
		Instance i2 = Instances.doubleInstance(i);
		investigateInstance(i);
		investigateInstance(i2);
		*/
		
		//breakIBHP();
		
		/*
		Instance i = new InstanceImpl(new int[][] {
			{0, 1, I, I},
			{0, 1, I, I},
			{I, 0, I, I},
			{I, I, I, I},
		});
		investigateInstance(i);
		*/
		
		//testTheBox();
		
		/*
		hashproduct.popmatch.threesat.Instance tsi = randomThreesat(1, 1, rnd);
			//new hashproduct.popmatch.threesat.InstanceImpl(1, new int[][] {{0, 0, 0}});
		System.out.println(tsi.toString());
		Instance gpmi = new TransformToPopMatch(tsi).gpmi();
		
		Matching m = new MatchingImpl(gpmi, new int[] {
			 0,  4,  5,  6,  7,  8, -1,
			13, 10, 11, 12, 14,  1, 15,
			 9, 16, 17, 18, 19, 20, 21,
		});
		MatchingAnalysis ma = Pressurizer.pressurize(m);
		System.out.println("-- Manually entered");
		System.out.println(ma);
		
		Matching cmmm = ClassMaximalMatcher.solve1(gpmi);
		MatchingAnalysis cmmma = Pressurizer.pressurize(cmmm);
		System.out.println("-- Class maximal matching");
		System.out.println(cmmma);
		*/
		
		//investigateThreesat();
		//breakIBHP();
	}
	
	private static void marginInvestigateInstance(Instance i) {
		Logger.out.append("Instance: " + Statics.toString(i) + "\n");
		//Solver[] solvers = Solvers.marginSolvers;
		Solver[] solvers = new Solver[] {
			new BruteForceSolver(MarginComparator.it),
			//MonotonicMarginVotingSolver.it,
			//ClassMaximalMatcher.it
			CompositeMarginSolver.it
			};
		MarginAnalysis[] mas = new MarginAnalysis[solvers.length];
		
		int bestUm = Integer.MAX_VALUE;
		for (int s = 0; s < solvers.length; s++) {
			mas[s] = Marginizer.marginize(solvers[s].solve(i));
			bestUm = min(bestUm, mas[s].unpopularity);
		}
		//boolean loses = false, wins = false;
		for (int s = 0; s < solvers.length; s++) {
			/*if (mas[s].unpopularity != bestUm)
				if (solvers[s] == ClassMaximalMatcher.it)
					loses = true;
				else
					wins = true;*/
			Logger.out.append("-- " + solvers[s].solverName()
				+ (mas[s].unpopularity == bestUm ? "" : " (SUBOPTIMAL)") + "\n");
			Logger.out.append(mas[s].toString());
			Logger.out.append('\n');
		}
		
		/*
		if (loses)
			System.out.println("*** CMM LOSES ***");
		//if (loses || wins)
			System.out.println(sb.toString());
			*/
		//return bestUm;
		if (mas[1].unpopularity > mas[0].unpopularity)
			Logger.isInteresting = true;
		Logger.endCase();
	}

	private static void marginReductionStuff1() {
		Instance i = new InstanceImpl(new int[][] {
			{I, 1, 2, I, I, I, I, I, I}, // left contender 1
			{I, 1, 2, I, I, I, I, I, I}, // left contender 2
			{I, 1, 2, 0, I, I, I, I, I}, // left linker 1
			{I, 1, 2, I, I, I, I, I, I}, // left linker 2
			{0, I, 1, I, I, I, I, I, I}, // left liker
			{I, I, I, I, I, 1, 2, I, I}, // right contender 1
			{I, I, I, I, I, 1, 2, I, I}, // right contender 2
			{I, I, I, I, I, 1, 2, 0, I}, // right linker 1
			{I, I, I, I, I, 1, 2, I, I}, // right linker 2
			{0, I, I, I, I, I, 1, I, I}, // left liker
		});
		//Matching m = MarginVotingSolver.it.solve(i);
		Matching m = new BruteForceSolver(MarginComparator.it).solve(i);
		MarginAnalysis ma = Marginizer.marginize(m);
		MarginAnalysis ma2 = Marginizer.marginize(ma.better);
		System.out.println(ma.toString());
		System.out.println(ma2.toString());
	}
	
	private static final boolean F = false, T = true;
	
	private static void marginReductionStuff2() {
		BruteForceSolver bfs = new BruteForceSolver(MarginComparator.it);
		Instance i = Instances.mbox;
		for (int s = 0; s < (1 << 6); s++) {
			int[] postsFor = new int[15];
			for (int p = 0; p < 15; p++)
				postsFor[p] = (p < 6 && (s & (1 << p)) != 0)
					? 4+p : lastResortFor(p);
			Matching m1 = new MatchingImpl(i, postsFor);
			Matching m = bfs.solveFromMatching(m1, 6);
			MarginAnalysis ma = Marginizer.marginize(m);
			String ss = "";
			for (int p = 0; p < 6; p++)
				ss += (s & (1 << p)) != 0 ? "t" : "a";
			System.out.println(ss + " => " + ma.unpopularity);
			//System.out.println(ma.toString());
		}
		/*
		Matching m1 = new MatchingImpl(i,
			new int[] {4, 5, ~2, 7, ~4, 9, ~6, ~7, ~8, ~9, ~10, ~11, ~12, ~13, ~14});
		Matching m = bfs.solveFromMatching(m1, 6);
		MarginAnalysis ma = Marginizer.marginize(m);
		System.out.println(ma.toString());
		*/
	}
	
	private static void miscellany() {
		/*
		hashproduct.popmatch.threesat.Instance tsi = //randomThreesat(1, 1, rnd);
			new hashproduct.popmatch.threesat.InstanceImpl(1, new int[][] {{0, 0, 0}});
		System.out.println(tsi.toString());
		Instance gpmi = new TransformToPopMatch(tsi).gpmi();
		// I realize the transformation makes the most sense for the factor
		// version, but it is a source of interesting instances.
		marginInvestigateInstance(gpmi);
		*/
		
		/*
		int ibhpLuf = 5;
		Instance i = Instances.ibhpFoo(ibhpLuf);
		System.out.println(i);
		*/
		
		/*
		IterativeSolverHelper.cycleDoesNotContainBestInteresting = true;
		// I want to see the first one to make sure stuff is working.
		//Logger.isInteresting = true;
		marginInvestigateRandomInstances(10000, 6, 4, 3, rnd);
		*/
	}
	
	public static void main(String[] args) {
		//marginInvestigateRandomInstances(1000, 7, 5, 5, rnd);
		//marginReductionStuff2();
		//threesatDemo();
		
		testReduction();
		
		// set a breakpoint here so you can inspect the results
		//System.out.println("hi");
	}
}
