package hashproduct.popmatch;

import java.util.*;
import static hashproduct.popmatch.Statics.*;

public class BruteForceSolver implements Solver {
	private final Comparator<Matching> comparator;
	
	public BruteForceSolver(Comparator<Matching> comparator) {
		this.comparator = comparator;
	}
	
	private Instance i;
	private Matching mbest;
	private MatchingImpl mbuf;
	
	private boolean isParetoEfficient(Matching m) {
		return m.equals(MatchingImproverUtils.makeParetoEfficient(m));
	}
	
	private void recSolve(int person) {
		if (person == i.numPeople()) {
			// Offer matching
			Matching m = new MatchingImpl(i, mbuf.postsFor); // clone
			
			if (mbest == null ||
							// This doesn't consider ties and thus doesn't catch
							// all Pareto inefficient matchings, but it should still help.
							isParetoEfficient(m) &&
							comparator.compare(mbuf, mbest) < 0
							)
				mbest = m;
		} else {
			recSolve(person + 1);
			for (int post = 0; post < i.numPosts(); post++) {
				if (mbuf.peopleFor[post] == NOBODY
								&& i.rankFor(person, post) != IMPOSSIBLE_RANK) {
					mbuf.peopleFor[post] = person;
					mbuf.postsFor[person] = post;
					recSolve(person + 1);
					mbuf.peopleFor[post] = NOBODY;
					mbuf.postsFor[person] = lastResortFor(person);
				}
			}
		}
	}
	
	@Override public Matching solve(Instance i1) {
		i = i1;
		mbuf = new MatchingImpl(i1);
		mbest = null;
		recSolve(0);
		return mbest;
	}

	public Matching solveFromMatching(Matching m1, int start) {
		i = m1.instance();
		mbuf = new MatchingImpl(i);
		for (int person = 0; person < start; person++)
			if (m1.postFor(person) != lastResortFor(person)) {
				mbuf.postsFor[person] = m1.postFor(person);
				mbuf.peopleFor[m1.postFor(person)] = person;
			}
		mbest = m1;
		recSolve(start);
		return mbest;
	}

	@Override public String solverName() {
		// TODO Identify the comparator.
		return "Brute-force solver";
	}

}
