package hashproduct.popmatch;
import static hashproduct.popmatch.Statics.*;

public final class InstanceImpl implements Instance {
	public int numPeople;
	public int numPosts;
	
	public int[][] ranks;
	public int[] numRanks;
	
	public InstanceImpl(int[][] ranks) {
		this(ranks.length, ranks[0].length, ranks);
	}

	public InstanceImpl(int numPeople, int numPosts, int[][] ranks) {
		this.numPeople = numPeople;
		this.numPosts = numPosts;
		this.ranks = ranks;
		numRanks = new int[numPeople];
		for (int person = 0; person < numPeople; person++) {
			int numRanks1 = 0;
			for (int post = 0; post < numPosts; post++)
				numRanks1 = Math.max(numRanks1, ranks[person][post] + 1);
			numRanks[person] = numRanks1;
		}
	}

	public int numPeople() {
		return numPeople;
	}

	public int numPosts() {
		return numPosts;
	}

	public int rankFor(int person, int post) {
		if (isALastResort(post))
			if (ownerOfLastResort(post) == person)
				return LAST_RESORT_RANK;
			else
				return IMPOSSIBLE_RANK;
		else
			return ranks[person][post];
	}

	public int numRanksFor(int person) {
		return numRanks[person];
	}
	
	@Override public boolean equals(Object o) {
		return (o instanceof Instance) && equal(this, (Instance) o); 
	}
	
	@Override public int hashCode() {
		return Statics.hashCode(this);
	}
	
	@Override public String toString() {
		return Statics.toString(this);
	}
}
