package hashproduct.popmatch;
import static hashproduct.popmatch.Statics.*;
import static java.util.Arrays.*;

import java.util.*;

public class Instances {
	private static final int I = IMPOSSIBLE_RANK;
	
	private Instances() {}
	
	public static final Instance intelBoardInstance =
		new InstanceImpl(new int[][] {
			{0, 1, 2, 3},
			{0, 1, 2, 3},
			{0, 1, 2, 3},
			{0, 1, 3, 2},
		});
	
	/*
	 *	{
	 *		{0, 1, 2, 3, 4},
	 *		{0, 1, 2, 3, 4},
	 *		{0, 1, 2, 3, 4},
	 *		{0, 1, 2, 3, 4},
	 *		{0, 1, 2, 3, 4},
	 *	}
	 */
	public static Instance contention(int luf) {
		int numP = luf + 1;
		int[] ranks1 = new int[numP];
		for (int i = 0; i < numP; i++)
			ranks1[i] = i;
		int[][] ranks = new int[numP][];
		for (int i = 0; i < numP; i++)
			ranks[i] = ranks1.clone();
		return new InstanceImpl(numP, numP, ranks);
	}
	
	/*
	 *	{
	 *		{1, I, I, I, I, I, I, 2},
	 *		{0, 1, I, I, I, I, I, 2},
	 *		{I, 0, 1, I, I, I, I, 2},
	 *		{I, I, 0, 1, I, I, I, 2},
	 *		{I, I, I, 0, 1, I, I, 2},
	 *		{I, I, I, I, 0, 1, I, 2},
	 *		{I, I, I, I, I, 0, 1, 2},
	 *		{I, I, I, I, I, I, 0, I},
	 *		{I, I, I, I, I, I, I, 0},
	 *	}
	 */
	public static Instance iprHard(int iprLuf) {
		int numPeople = iprLuf + 2;
		int numPosts = iprLuf + 1;
		int[][] ranks = new int[numPeople][numPosts];
		for (int person = 0; person < numPeople; person++) {
			fill(ranks[person], IMPOSSIBLE_RANK);
			if (person == iprLuf + 1)
				ranks[person][iprLuf] = 0;
			else {
				if (person < iprLuf) {
					ranks[person][person] = 1;
					ranks[person][iprLuf] = 2;
				}
				if (person > 0)
					ranks[person][person - 1] = 0;
			}
		}
		return new InstanceImpl(numPeople, numPosts, ranks);
	}
	
	public static Instance ibhpFoo(int ibhpLuf) {
		int x = ibhpLuf + 1, y = ibhpLuf - 1;
		int numPeople = x + y;
		int numPosts = x + y;
		int[] initRanks = new int[numPosts];
		fill(initRanks, IMPOSSIBLE_RANK);
		int[][] ranks = new int[numPeople][];
		for (int person = 0; person < y; person++) {
			ranks[person] = initRanks.clone();
			ranks[person][person] = 0;
		}
		for (int post = 0; post < y; post++)
			initRanks[post] = 2;
		for (int pe2 = 0; pe2 < x; pe2++) {
			int person = y + pe2;
			ranks[person] = initRanks.clone();
			for (int po2 = 0; po2 < x; po2++)
				ranks[person][y + po2] = (po2 >= 2) ? po2 + 1 : po2;
		}
		return new InstanceImpl(numPeople, numPosts, ranks);
	}
	
	public static Matching ibhpFooMatch(int ibhpLuf) {
		Instance i = ibhpFoo(ibhpLuf);
		int[] postsFor = new int[2 * ibhpLuf];
		for (int j = 0; j < ibhpLuf - 1; j++)
			postsFor[j] = lastResortFor(j);
		for (int j = ibhpLuf - 1; j < 2 * ibhpLuf; j++)
			postsFor[j] = j - (ibhpLuf - 1);
		return new MatchingImpl(i, postsFor);
	}
	
	// Algorithm doesn't seem to work. -- Matt 2006.09.03
	public static Instance doubleInstance(Instance i) {
		if (hasTies(i))
			throw new IllegalArgumentException("ties");
		int numPeople2 = 2 * i.numPeople();
		int numPosts2 = 2 * i.numPosts() + i.numPeople();
		int[][] ranks2 = new int[numPeople2][numPosts2];
		for (int j = 0; j < numPeople2; j++)
			fill(ranks2[j], IMPOSSIBLE_RANK);
		for (int person = 0; person < i.numPeople(); person++) {
			for (int person2 = 2 * person; person2 < 2 * (person + 1); person2++) {
				// Translate the ordinary posts
				for (int post = 0; post < i.numPosts(); post++) {
					int rank = i.rankFor(person, post);
					if (rank != IMPOSSIBLE_RANK)
						for (int j = 0; j < 2; j++)
							ranks2[person2][2 * post + j] = 2 * rank + j;
				}
				// Translate the last resort primary
				int lrp = 2 * i.numPosts() + person;
				ranks2[person2][lrp] = 2 * i.numRanksFor(person);
			}
		}
		return new InstanceImpl(numPeople2, numPosts2, ranks2);
	}
	
	public static Instance randomInstance(int numPeople, int numPosts, int numRanks, Random r) {
		int[][] ranks = new int[numPeople][numPosts];
		for (int person = 0; person < numPeople; person++)
			for (int post = 0; post < numPosts; post++) {
				int rank = r.nextInt(numRanks + 1);
				if (rank == numRanks)
					rank = IMPOSSIBLE_RANK;
				ranks[person][post] = rank;
			}
		return new InstanceImpl(numPeople, numPosts, ranks);
	}
	
	// This doesn't accomplish what we want.
	public static final Instance mbox = new InstanceImpl(new int[][] {
		{I, I, I, I, 1, I, I, I, I, I},
		{I, I, I, I, I, 1, I, I, I, I},
		{I, I, I, I, I, I, 1, I, I, I},
		{I, I, I, I, I, I, I, 1, I, I},
		{I, I, I, I, I, I, I, I, 1, I},
		{I, I, I, I, I, I, I, I, I, 1},
		{2, I, 3, I, 1, I, I, I, I, I},
		{2, I, 3, I, I, 1, I, I, I, I},
		{2, I, 3, I, I, I, 1, I, I, I},
		{2, I, I, 3, I, I, I, 1, I, I},
		{2, I, I, 3, I, I, I, I, 1, I},
		{2, I, I, 3, I, I, I, I, I, 1},
		{2, 1, I, I, I, I, I, I, I, I},
		{2, 1, I, I, I, I, I, I, I, I},
		{2, 1, I, I, I, I, I, I, I, I},
	});
}
