package hashproduct.popmatch;

import java.util.*;

public class IterativeSolverHelper {
	private IterativeSolverHelper() {}
	
	private static void explainImprovement1(MatchingImprover improver,
		Instance i, Matching m, Matching m2, StringBuilder out) {
		if (improver instanceof ExplainingMatchingImprover)
			((ExplainingMatchingImprover) improver).explainImprovement(m, out);
		else
			Statics.format(i, m, m2, null, out);
	}
	
	public static boolean cycleDoesNotContainBestInteresting;
	
	public static Matching iterateToCycle(MatchingImprover improver,
		Comparator<Matching> comparator, Matching m) {
		Matching mBest = m;
		// Remember what led to what so that we can reconstruct the cycle we end in.
		Map<Matching, Matching> considered = new LinkedHashMap<Matching, Matching>();
		while (!considered.containsKey(m)) {
			Matching m2 = improver.improve(m);
			if (m2 == null)
				m2 = m;
			considered.put(m, m2);
			m = m2;
			if (comparator.compare(mBest, m) >= 0)
				// Take m if it's at least as good as the previous best.
				// Convenient for seeing if a strictly better matching came before the cycle.
				mBest = m;
		}
		{
			boolean sawBest = false;
			Matching m1 = m;
			Logger.out.append("-- IterativeSolverHelper closed the cycle.  Here it is: --\n");
			do {
				if (m.equals(mBest)) {
					Logger.out.append("---- Best matching\n");
					sawBest = true;
				}
				Matching m2 = considered.get(m);
				explainImprovement1(improver, m.instance(), m, m2, Logger.out);
				m = m2;
			} while (!m.equals(m1));
			Logger.out.append("-- End of cycle --\n");
			if (!sawBest) {
				Logger.isInteresting |= cycleDoesNotContainBestInteresting;
				Logger.out.append("Cycle did not contain the best matching, which is:\n");
				explainImprovement1(improver, m.instance(), mBest, considered.get(mBest), Logger.out);
			}
		}
		// we're in a cycle, presumably of "best" matchings
		// now make it Pareto efficient
		return MatchingImproverUtils.makeParetoEfficient(m);
	}
	
	// Comparator is used to choose the "best" matching encountered during
	// the iteration.  "Less" means a better matching.
	public static Matching iterateToCycle(MatchingImprover improver,
			Comparator<Matching> comparator, Instance i) {
		// Start with an arbitrary Pareto efficient matching
		Matching m = MatchingImproverUtils.paretoEfficientMatchingFor(i);
		return iterateToCycle(improver, comparator, m);
	}
	
	public static Matching iterateMonotonic(MultipleMatchingImprover improver,
		Comparator<Matching> comparator, Matching m) {
		outer: for (;;) {
			Iterator<Matching> itr = improver.improvementIteratorFor(m);
			while (itr.hasNext()) {
				Matching m2 = itr.next();
				if (comparator.compare(m2, m) < 0) {
					m = m2;
					continue outer;
				}
			}
			// Could not improve.
			return m;
		}
	}

	public static Matching iterateMonotonic(MultipleMatchingImprover improver,
		Comparator<Matching> comparator, Instance i) {
		// Start with an arbitrary Pareto efficient matching
		Matching m = MatchingImproverUtils.paretoEfficientMatchingFor(i);
		return iterateMonotonic(improver, comparator, m);
	}
}
