package hashproduct.popmatch;

import static hashproduct.popmatch.Statics.*;

import java.util.*;

import hashproduct.popmatch.util.*;

// LUMM -- UNFINISHED

public class Marginizer {
	private Marginizer() {}
	
	static class OurNode extends MinCostFlowSolver.Node<OurNode> {
		final boolean isPerson;
		final int num;
		OurNode(boolean isPerson, int num) {
			this.isPerson = isPerson;
			this.num = num;
		}
		
		// DEBUG
		@Override public String toString() {
			return (num == -1 ? "sourceSink" :
					(isPerson ? "person" : "post") + " " + num);
		}
	}
	
	private static int vote(int r1, int r2) {
		return (r2 < r1) ? 1 : (r2 > r1) ? -1 : 0;
	}
	
	// Helps us handle "matched edges"
	private static void addEdge1(MinCostFlowSolver<OurNode> mcfs,
		OurNode start, OurNode end, boolean reverse, int value) {
		if (reverse)
			mcfs.addEdge(end, start, -value);
		else
			mcfs.addEdge(start, end, value);
	}
	
	/*
	 * Really, we're doing "min-cost bipartite matching".  We initialize edge
	 * directions to reflect the initial matching.  Alternatively, we could
	 * make all the edges forward, but the algorithm might be slower and (more
	 * importantly) we'd have to initialize totalValue to the negative of the
	 * number of people matched to non-last-resort posts, which is ugly.
	 * 
	 * sourcesink -> person1 -> post1 -----\
	 *                   \----> post2 -> sourcesink
	 * 
	 * To match someone to his/her last resort, send him/her back to sourcesink.
	 */
	
	public static MarginAnalysis marginize(Matching m) {
		Instance i = m.instance();
		// Make our solver.
		MinCostFlowSolver<OurNode> mcfs = new MinCostFlowSolver<OurNode>();

		// Node used for ends of alternating paths, including last resorts.
		OurNode sourceSink = new OurNode(false, -1);
		mcfs.nodes.add(sourceSink);
		// Introduce a node for each post.
		OurNode[] postNodes = new OurNode[i.numPosts()];
		for (int post = 0; post < i.numPosts(); post++) {
			postNodes[post] = new OurNode(false, post);
			mcfs.nodes.add(postNodes[post]);
			addEdge1(mcfs, postNodes[post], sourceSink, (m.personFor(post) != NOBODY), 0);
		}
		// Introduce a person-node and preference edges for each person.
		OurNode[] personNodes = new OurNode[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++) {
			int assignedPost = m.postFor(person);
			int assignedRank = i.rankFor(person, assignedPost);
			personNodes[person] = new OurNode(true, person);
			mcfs.nodes.add(personNodes[person]);
			for (int otherPost = 0; otherPost < i.numPosts(); otherPost++)
				if (i.rankFor(person, otherPost) != IMPOSSIBLE_RANK)
					addEdge1(mcfs, personNodes[person], postNodes[otherPost],
						(otherPost == assignedPost),
						vote(assignedRank, i.rankFor(person, otherPost)));
			// sourceSink behaves like a post node for the last resort.
			addEdge1(mcfs, personNodes[person], sourceSink,
				isALastResort(assignedPost),
				vote(assignedRank, LAST_RESORT_RANK));
		}
		// Solve the min-cost flow.
		mcfs.solve();
		// Construct the "better" matching.
		int[] postsFor2 = new int[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++) {
			OurNode pern = personNodes[person];
			boolean set = false;
			for (Map.Entry<OurNode, MinCostFlowSolver.Edge> en
							: pern.edges.entrySet()) {
				// The lack of capacity to match this person to a post indicates
				// that the person *is* already matched to the post.
				if (!en.getValue().capacity) {
					assert !set;
					OurNode postn = en.getKey();
					assert !postn.isPerson;
					postsFor2[person] = (postn == sourceSink) ?
						lastResortFor(person) : postn.num;
					set = true;
				}
			}
			assert set;
		}
		Matching m2 = new MatchingImpl(i, postsFor2);
		// mcfs.totalValue is the total vote in favor of the new matching.
		return new MarginAnalysis(m, mcfs.totalValue, m2);
	}
}
