package hashproduct.popmatch;
import static java.util.Arrays.*;
import static hashproduct.popmatch.Statics.*;

public final class MatchingImpl implements Matching {
	public Instance instance;
	public int[] peopleFor;
	public int[] postsFor;
	
	public MatchingImpl(Instance instance, int[] postsFor) {
		this.instance = instance;
		this.postsFor = postsFor.clone();
		for (int i = 0; i < instance.numPeople(); i++)
			if (this.postsFor[i] < 0)
				this.postsFor[i] = lastResortFor(i);
		peopleFor = new int[instance.numPosts()];
		fill(peopleFor, NOBODY);
		for (int i = 0; i < instance.numPeople(); i++)
			if (!isALastResort(postsFor[i])) {
				if (instance.rankFor(i, postsFor[i]) == IMPOSSIBLE_RANK)
					throw new IllegalArgumentException("Impossible to put person " + i + " in post " + postsFor[i] + ", instance: " + instance.toString());
				if (peopleFor[postsFor[i]] != NOBODY)
					throw new IllegalArgumentException("People " + peopleFor[postsFor[i]] +
						" and " + i + " both in post " + postsFor[i] + ", instance: " + instance.toString());
				peopleFor[postsFor[i]] = i;
			}
	}
	
	// Everyone to last resort; a basis for mutation
	public MatchingImpl(Instance instance) {
		this.instance = instance;
		peopleFor = new int[instance.numPosts()];
		fill(peopleFor, NOBODY);
		postsFor = new int[instance.numPeople()];
		for (int i = 0; i < instance.numPeople(); i++)
			postsFor[i] = lastResortFor(i);
	}

	public Instance instance() {
		return instance;
	}

	public int personFor(int post) {
		if (isALastResort(post)) {
			int o = ownerOfLastResort(post);
			if (postsFor[o] == post)
				return o;
			else
				return NOBODY;
		} else
			return peopleFor[post];
	}

	public int postFor(int person) {
		return postsFor[person];
	}

	@Override public boolean equals(Object o) {
		return (o instanceof Matching) && equal(this, (Matching) o);
	}
	
	@Override public int hashCode() {
		return Statics.hashCode(this);
	}
	
	@Override public String toString() {
		return Statics.toString(this);
	}
}
