package hashproduct.popmatch;
import static hashproduct.popmatch.Statics.*;

import java.util.*;

public class MatchingImproverUtils {
	private MatchingImproverUtils() {}
	
	/* Apply the change represented by the linked list of nodes to the matching.
	 * 
	 * what could we be given?
	 *
	 * pareto chain: vacant post, post of person filling it, etc.
	 * 
	 * pareto cycle: post of person cycled to end, post of person filling it, etc.
	 * 
	 * knock: post being knocked out, post of person filling it, etc.
	 * */
	public static Matching apply(Matching m, Pressurizer.PressureNode start,
		boolean keepParetoEfficient) {
		Instance i = m.instance();
		int[] newPostsFor = new int[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++)
			newPostsFor[person] = m.postFor(person);
		
		int startPerson = m.personFor(start.post);
		if (startPerson != NOBODY)
			newPostsFor[startPerson] = lastResortFor(startPerson);
		
		// If we have a path leading into a cycle, we want to ignore the cycle.
		Set<Pressurizer.PressureNode> seen = new LinkedHashSet<Pressurizer.PressureNode>();
		Pressurizer.PressureNode n = start;
		while (!seen.contains(n)) {
			if (n == null)
				break;
			seen.add(n);
			n = n.from;
		}
		Pressurizer.PressureNode stopAt = n;
		
		n = start;
		while (n.from != stopAt) {
			newPostsFor[m.personFor(n.from.post)] = n.post;
			if (n.from == start)
				// ok, we went around once and reinstated the start person
				break;
			n = n.from;
		}
		
		Matching m2 = new MatchingImpl(i, newPostsFor);
		
		if (keepParetoEfficient && startPerson != NOBODY
						&& m2.postFor(startPerson) == lastResortFor(startPerson))
			m2 = paretoPromote(m2, startPerson);
		
		return m2;
	}
	
	public static void storeImprovement(PressureAnalysis ma) {
		ma.better = apply(ma.m, ma.n, true); // for the moment
	}
	
	// FIX: Doesn't consider ties
	public static Matching paretoPromote(Matching m, int promotePerson) {
		Instance i = m.instance();
		int bestPost = m.postFor(promotePerson);
		for (int post = 0; post < i.numPosts(); post++) {
			if (m.personFor(post) == NOBODY
							&& i.rankFor(promotePerson, post) < i.rankFor(promotePerson, bestPost))
				bestPost = post;
		}
		int[] newPostsFor = new int[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++)
			newPostsFor[person] = m.postFor(person);
		newPostsFor[promotePerson] = bestPost;
		return new MatchingImpl(i, newPostsFor);
	}
	
	public static Matching makeParetoEfficient(Matching m) {
		Instance i = m.instance();
		for (int person = 0; person < i.numPeople(); person++)
			m = paretoPromote(m, person);
		return m;
	}
	
	public static Matching paretoEfficientMatchingFor(Instance i) {
		return makeParetoEfficient(everyoneToLastResort(i));
	}
}
