package hashproduct.popmatch;

public class MonotonicPressureBumpingSolver extends AbstractSolver
		implements MatchingImprover {
	private MonotonicPressureBumpingSolver() {
		super("Monotonic pressure-bumping solver");
	}
	public static final MonotonicPressureBumpingSolver it = new MonotonicPressureBumpingSolver();
	
	public Matching solve(Instance i) {
		return IterativeSolverHelper.iterateToCycle(this, FactorComparator.it, i);
	}
	
	public Matching improve(Matching m) {
		Instance i = m.instance();
		PressureAnalysis ma = Pressurizer.pressurize(m);
		for (int post = 0; post < i.numPosts(); post++) {
			Matching m2 = MatchingImproverUtils.apply(m, ma.nodes[post], true);
			PressureAnalysis ma2 = Pressurizer.pressurize(m2);
			if (ma2.unpopularity < ma.unpopularity) {
				return m2;
			}
		}
		// We tried all the posts without finding a way to reduce the
		// unpopularity factor.
		return null;
	}
}
