package hashproduct.popmatch;

import static hashproduct.popmatch.Statics.*;
import hashproduct.popmatch.Pressurizer.*;
import hashproduct.popmatch.util.*;

public class PressureAnalysis {
	
	/* The analyzed matching */
	public Matching m;
	
	/* The unpopularity of the matching, i.e., the highest pressure,
	 * or PARETO_INEFFICIENT. */
	public int unpopularity;
	
	/* This classification contains the actual pressures of the matching,
	 * but it is null if the matching has a cycle Pareto improvement.
	 * If the matching is Pareto efficient, this classification stabilizes it. */
	public Pressurization c;
	
	/* If the matching is Pareto efficient, this is a node with the
	 * highest pressure (say x), and it is the start of a linked list of
	 * x+1 InstanceEdgeNodes (linked by from field) that tell you the
	 * improvement with unpopularity factor x.
	 * 
	 * If the matching is not Pareto efficient, this is the start of a
	 * linked list of InstanceEdgeNodes (linked by from field) that tell
	 * you the Pareto improvement.  There are two possibilities:
	 * 
	 * (1) The linked list begins with a vacant-post node and ends with a
	 *     node that has from == null.
	 * 
	 * (2) The linked list is a cycle. */
	public PressureNode n;
	
	/* Other analysis information; possibly useful; access discouraged */
	public PressureNode[] nodes;
	public PressureNode[] lastResortNodes;
	public BellmanFordSolver<PressureNode> bfs;
	
	/* An improved matching; may be filled by MatchingImprover */
	public Matching better;
	
	PressureAnalysis(Matching m, int unpopularity, Pressurization c, PressureNode n,
		PressureNode[] nodes, PressureNode[] lastResortNodes, BellmanFordSolver<PressureNode> bfs) {
		this.m = m;
		this.unpopularity = unpopularity;
		this.c = c;
		this.n = n;
		this.nodes = nodes;
		this.lastResortNodes = lastResortNodes;
		this.bfs = bfs;
	}
	
	@Override public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("Unpopularity factor: ");
		sb.append(unpopularity == PARETO_INEFFICIENT
			? "Pareto inefficient" : unpopularity);
		sb.append('\n');
		format(m.instance(), m, null, c, sb);
		return sb.toString();
	}
}
