package hashproduct.popmatch;

public class PressureReducingSolver extends AbstractSolver
		implements MatchingImprover {
	private PressureReducingSolver() {
		super("Pressure-reducing solver");
	}
	public static final PressureReducingSolver it = new PressureReducingSolver();
	
	public Matching improve(Matching m) {
		Instance i = m.instance();
		Pressurization c = Pressurizer.pressurize(m).c;
		// Try to reduce pressure on each post.
		for (int post = 0; post < i.numPosts(); post++) {
			if (c.pressureFor(post) == 0)
				continue;
			int[] newPressures = new int[i.numPosts()];
			for (int p = 0; p < i.numPosts(); p++)
				newPressures[p] = (p == post)
					? c.pressureFor(p) - 1 : c.pressureFor(p);
			Pressurization c2 = new PressurizationImpl(i, newPressures);
			// Is there a matching for the reduced classification?
			Matching m2 = PressurizationMatcher.matchingFor(c2);
			if (m2 != null) {
				// We made an improvement; start over looking for a new
				// post on which to decrease pressure
				return m2;
			}
		}
		// we tried all posts without making an improvement
		return null;
	}
	
	public Matching solve(Instance i) {
		return IterativeSolverHelper.iterateToCycle(this, FactorComparator.it, i);
	}
}
