package hashproduct.popmatch;
import java.util.*;
import hashproduct.popmatch.util.BipartiteMatcher;
import static hashproduct.popmatch.Statics.*;

public class PressurizationMatcher {
	private PressurizationMatcher() {}
	
	// When we use a BipartiteMatcher, foos are people and bars are posts;
	// posts numbered numPosts through numPosts + numPeople - 1 are last resorts
	
	// ClassMaximalSolver uses this now...
	static Matching toMatching(Instance i, BipartiteMatcher bm) {
		int[] postsFor = new int[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++) {
			postsFor[person] =
				(bm.barFor[person] >= i.numPosts() ||
								bm.barFor[person] == BipartiteMatcher.UNMATCHED)
				? lastResortFor(person)
				: bm.barFor[person];
		}
		return new MatchingImpl(i, postsFor);
	}
	
	static class PreferenceClass extends ArrayList<Integer> {}
	
	public static Matching matchingFor(Pressurization c) {
		Instance i = c.instance();
		BipartiteMatcher bm = new BipartiteMatcher(i.numPeople(),
			i.numPosts() + i.numPeople());
		
		// Tell the bipartite matcher what's allowed
		for (int person = 0; person < i.numPeople(); person++) {
			
			// Read the whole preference list
			PreferenceClass[] pclasses = new PreferenceClass
				[i.numRanksFor(person) + 1 /* last resort */];
			for (int pci = 0; pci < pclasses.length; pci++)
				pclasses[pci] = new PreferenceClass();
			for (int post = 0; post < i.numPosts(); post++) {
				int rank = i.rankFor(person, post);
				if (rank != IMPOSSIBLE_RANK)
					pclasses[rank].add(post);
			}
			pclasses[pclasses.length - 1].add(lastResortFor(person));
			
			// Scan the preference list a class at a time, determining which
			// posts have a lower pressure than all better posts.
			int minPressure = Integer.MAX_VALUE;
			for (int pci = 0; pci < pclasses.length; pci++) {
				int newMinPressure = minPressure;
				for (int post : pclasses[pci])
					newMinPressure = Math.min(newMinPressure, c.pressureFor(post));
				for (int post : pclasses[pci])
					if (c.pressureFor(post) < minPressure && c.pressureFor(post) <= newMinPressure)
						bm.ok[person][isALastResort(post)
							? i.numPosts() + ownerOfLastResort(post) : post] = true;
				minPressure = newMinPressure;
			}
		}
		
		// Tell the matcher that nonzero-pressure posts must be filled
		for (int post = 0; post < i.numPosts(); post++)
			if (c.pressureFor(post) > 0)
				bm.barRequired[post] = true;
		
		boolean reqdBarsFilled = bm.augment();
		if (reqdBarsFilled && bm.numMatched == i.numPeople())
			return toMatching(i, bm);
		else
			return null;
	}
}
