package hashproduct.popmatch;

import hashproduct.popmatch.util.*;

public abstract class Statics {
	private Statics() {}
	
	public static final int LAST_RESORT_RANK = Integer.MAX_VALUE - 1;
	public static final int IMPOSSIBLE_RANK = Integer.MAX_VALUE;
	
	// Matching.personFor
	public static int NOBODY = -1;
	
	// MatchingAnalysis.unpopularity
	public static int PARETO_INEFFICIENT = Integer.MAX_VALUE;
	
	public static boolean isALastResort(int post) {
		return post < 0;
	}
	
	public static int ownerOfLastResort(int post) {
		return ~post;
	}
	
	public static int lastResortFor(int person) {
		return ~person;
	}
	
	public static boolean equal(Instance i1, Instance i2) {
		if (i1.numPeople() != i2.numPeople() || i1.numPosts() != i2.numPosts())
			return false;
		for (int person = 0; person < i1.numPeople(); person++)
			for (int post = 0; post < i1.numPosts(); post++)
				if (i1.rankFor(person, post) != i2.rankFor(person, post))
					return false;
		return true;
	}
	
	public static int hashCode(Instance i) {
		Hasher h = new Hasher();
		h.mash(i.numPeople());
		h.mash(i.numPosts());
		for (int person = 0; person < i.numPeople(); person++)
			for (int post = 0; post < i.numPosts(); post++)
				h.mash(i.rankFor(person, post));
		return h.hash;
	}
	
	public static String toString(Instance i) {
		StringBuilder sb = new StringBuilder();
		format(i, null, null, null, sb);
		return sb.toString();
	}
	
	// Assume they belong to the same instance
	public static boolean equal(Matching m1, Matching m2) {
		Instance i = m1.instance();
		for (int person = 0; person < i.numPeople(); person++)
			if (m1.postFor(person) != m2.postFor(person))
				return false;
		return true;
	}
	
	public static int hashCode(Matching m) {
		Hasher h = new Hasher();
		Instance i = m.instance();
		for (int person = 0; person < i.numPeople(); person++)
			h.mash(m.postFor(person));
		return h.hash;
	}
	
	public static String toString(Matching m) {
		StringBuilder sb = new StringBuilder();
		format(m.instance(), m, null, null, sb);
		return sb.toString();
	}
	
	// Assume they belong to the same instance
	public static boolean equal(Pressurization c1, Pressurization c2) {
		Instance i = c1.instance();
		for (int post = 0; post < i.numPosts(); post++)
			if (c1.pressureFor(post) != c2.pressureFor(post))
				return false;
		return true;
	}
	
	public static int hashCode(Pressurization c) {
		Hasher h = new Hasher();
		Instance i = c.instance();
		for (int post = 0; post < i.numPosts(); post++)
			h.mash(c.pressureFor(post));
		return h.hash;
	}
	
	public static String toString(Pressurization c) {
		StringBuilder sb = new StringBuilder();
		format(c.instance(), null, null, c, sb);
		return sb.toString();
	}
	
	// The instance is required.  The matching and the classification are
	// optional; if they are provided, they must be for the given instance.
	public static void format(Instance i, Matching m, Matching m2,
			Pressurization c, StringBuilder sb) {
		sb.append("{ " + i.numPeople() + " people (rows), " + i.numPosts() + " posts (cols)\n");
		sb.append("Rank matrix");
		boolean with = false;
		if (m != null) {
			sb.append(" with (matching)");
			with = true;
		}
		if (m2 != null) {
			sb.append(with ? " and" : " with");
			sb.append(" <alternate matching>");
			with = true;
		}
		if (c != null) {
			sb.append(with ? " and" : " with");
			sb.append(" pressures");
			with = true;
		}
		sb.append(":\n");
		int numRows = i.numPeople() + 1;
		int numCols =
				1 /* [ */ + 1 /* initial space */
				+ 2 * (i.numPosts() + 1) /* each post and following space */
				+ 1 /* ] */ + 1 /* space after */ + 1 /* change sign */;
		String[][] cells = new String[numRows][numCols];
		if (c != null) {
			String[] cCells = cells[0];
			for (int post = 0; post < i.numPosts(); post++) {
				cCells[2 + 2 * post] = String.valueOf(c.pressureFor(post));
				if (m != null &&
					c.pressureFor(post) > 0 && m.personFor(post) == NOBODY)
					cCells[2 + 2 * post + 1] = "!"; // Path Pareto inefficiency
			}
		}
		for (int person = 0; person < i.numPeople(); person++) {
			String[] rowCells = cells[person + 1];
			rowCells[0] = "[";
			rowCells[1] = " ";
			for (int post = 0; post < i.numPosts(); post++) {
				int rank = i.rankFor(person, post);
				rowCells[2 + 2 * post] = (rank == IMPOSSIBLE_RANK) ? "."
					: String.valueOf(i.rankFor(person, post));
				rowCells[2 + 2 * post + 1] = " ";
			}
			rowCells[numCols - 5] = "LR";
			rowCells[numCols - 4] = " ";
			rowCells[numCols - 3] = "]";
			if (m != null && m2 != null) {
				int deltaRank = i.rankFor(person, m2.postFor(person))
					- i.rankFor(person, m.postFor(person));
				rowCells[numCols - 2] = " ";
				// Remember, lower rank is happier
				rowCells[numCols - 1] = (deltaRank < 0) ? "+" : (deltaRank > 0) ? "-" : "";
			}
			if (m2 != null) {
				int matchPost = m2.postFor(person);
				int matchPost2 = (matchPost == lastResortFor(person)) ? i.numPosts()
					: matchPost;
				rowCells[2 + 2 * matchPost2 - 1] = "<";
				rowCells[2 + 2 * matchPost2 + 1] = ">";
			}
			if (m != null) {
				int matchPost = m.postFor(person);
				int matchPost2 = (matchPost == lastResortFor(person)) ? i.numPosts()
					: matchPost;
				rowCells[2 + 2 * matchPost2 - 1] = "(";
				rowCells[2 + 2 * matchPost2 + 1] = ")";
			}
		}
		Formatter.formatTable(numRows, numCols, cells, sb);
		sb.append("}");
	}
	
	public static Matching everyoneToLastResort(Instance i) {
		int[] postsFor = new int[i.numPeople()];
		for (int person = 0; person < i.numPeople(); person++)
			postsFor[person] = lastResortFor(person);
		return new MatchingImpl(i, postsFor);
	}
	
	public static boolean hasTies(Instance i) {
		for (int person = 0; person < i.numPeople(); person++) {
			boolean[] seenRank = new boolean[i.numRanksFor(person)];
			for (int post = 0; post < i.numPosts(); post++) {
				int rank = i.rankFor(person, post);
				if (rank != IMPOSSIBLE_RANK) {
					if (seenRank[rank])
						return true;
					seenRank[rank] = true;
				}
			}
		}
		return false;
	}
}
