package hashproduct.popmatch.threesat;

import hashproduct.popmatch.util.*;

public class Statics {
	private Statics() {}
	
	public static final boolean isNegated(int ref) {
		return ref < 0;
	}
	
	public static final int variableOf(int ref) {
		return isNegated(ref) ? ~ref : ref;
	}
	
	public static boolean equal(Instance i1, Instance i2) {
		if (i1.numVariables() != i2.numVariables()
						|| i1.numStatements() != i2.numStatements())
			return false;
		for (int s = 0; s < i1.numStatements(); s++)
			for (int p = 0; p < 3; p++)
				if (i1.referenceAt(s, p) != i2.referenceAt(s, p))
					return false;
		return true;
	}
	
	public static int hashCode(Instance i1) {
		Hasher h = new Hasher();
		h.mash(i1.numVariables());
		for (int s = 0; s < i1.numStatements(); s++)
			for (int p = 0; p < 3; p++)
				h.mash(i1.referenceAt(s, p));
		return h.hash;
	}
	
	public static String toString(Instance i) {
		StringBuilder sb = new StringBuilder();
		format(i, sb);
		return sb.toString();
	}
	
	public static void format(Instance i, StringBuilder sb) {
		sb.append("& {\n");
		for (int s = 0; s < i.numStatements(); s++) {
			sb.append("(");
			for (int p = 0; p < 3; p++) {
				if (p > 0)
					sb.append(" | ");
				int ref = i.referenceAt(s, p);
				sb.append(isNegated(ref) ? '~' : ' ');
				sb.append(variableOf(ref));
			}
			sb.append("),\n");
		}
		sb.append("}");
	}
	
	public static boolean satisfies(boolean[] vals, Instance i1) {
		assert vals.length == i1.numVariables();
		stmtLoop: for (int s = 0; s < i1.numStatements(); s++) {
			for (int p = 0; p < 3; p++) {
				int ref = i1.referenceAt(s, p);
				if (vals[variableOf(ref)] == !isNegated(ref))
					continue stmtLoop;
			}
			return false;
		}
		return true;
	}
	
	// This takes EXPONENTIAL TIME, of course!!!
	public static boolean[] solution(Instance i1) {
		boolean[] vals = new boolean[i1.numVariables()];
		if (satisfiable1(i1, vals, 0))
			return vals;
		else
			return null;
	}
	
	private static boolean satisfiable1(Instance i1, boolean[] vals, int j) {
		if (j == i1.numVariables())
			return satisfies(vals, i1);
		else {
			vals[j] = false;
			if (satisfiable1(i1, vals, j + 1))
				return true;
			vals[j] = true;
			if (satisfiable1(i1, vals, j + 1))
				return true;
			return false;
		}
	}
}
