package hashproduct.popmatch.util;
import java.util.*;

public final class BellmanFordSolver<N extends BellmanFordSolver.Node<N>> {
	
	// Extend me if you want to store additional information
	// so you can figure out what happened in case of a cycle.
	public static class Node<N extends Node<N>> {
		// Fill this.
		public Map<N, Integer> edges = new LinkedHashMap<N, Integer>();
		
		public final void putEdge(N dest, int len) {
			int oldLen = edges.containsKey(dest)
				? edges.get(dest): Integer.MAX_VALUE;
			if (len < oldLen)
				edges.put(dest, len);
		}
		
		// These are meaningful after you solve.
		public int dist = Integer.MAX_VALUE;
		public N from = null;
		
		// Used by solve
		boolean cycleChecked = false;
		
		public Node() {}
	}
	
	// Add to me!
	public Set<N> nodes = new LinkedHashSet<N>();
	
	private Set<N> toCheck = new LinkedHashSet<N>();
	
	// Returns null on success or a cycled node on failure.
	public N solve() {
		for (int pass = 0; !toCheck.isEmpty(); pass++) {
			if (pass >= nodes.size()) {
				// Ugh, some nodes might be merely reachable from the cycle
				// rather than part of it.  Find one that is part of it.
				// Issue arose in MarginVotingSolver on an instance made from a
				// 3SAT. 
				N n = toCheck.iterator().next();
				while (!n.cycleChecked) {
					n.cycleChecked = true;
					n = n.from;
				}
				return n;
			}
			Set<N> saveToCheck = toCheck;
			toCheck = new LinkedHashSet<N>();
			for (N n : saveToCheck)
				for (Map.Entry<N, Integer> e : n.edges.entrySet())
					offer(e.getKey(), n.dist + e.getValue(), n);
		}
		return null;
	}

	// Call me to enter your starting point(s)!
	public void offer(N n, int dist, N from) {
		if (dist < n.dist) {
			n.dist = dist;
			n.from = from;
			toCheck.add(n);
		}
	}
}
