package hashproduct.popmatch.util;
import static java.util.Arrays.*;

public final class BipartiteMatcher {
	// Gallai-Edmonds decomposition codes
	public enum GED {
		E, U, O
	}
	
	public static final int UNMATCHED = -1;
	
	public int numFoos, numBars;
	public boolean[][] ok; // set cells in me!
	public boolean[] barRequired; // set cells in me if desired
	
	// matching; you can modify these directly if you're careful
	public int[] barFor;
	public int[] fooFor;
	public int numMatched;
	
	// filled by decompose
	public GED[] fooGed;
	public GED[] barGed;
	
	public BipartiteMatcher(int numFoos, int numBars) {
		this.numFoos = numFoos;
		this.numBars = numBars;
		ok = new boolean[numFoos][numBars];
		barRequired = new boolean[numBars];
		barFor = new int[numFoos];
		fill(barFor, UNMATCHED);
		fooFor = new int[numBars];
		fill(fooFor, UNMATCHED);
		fooGed = new GED[numFoos];
		barGed = new GED[numBars];
	}
	
	private boolean[] fooTried;
	private boolean[] barTried;
	
	private boolean rematchFoo(int foo) {
		if (!fooTried[foo]) {
			fooTried[foo] = true;
			for (int bar = 0; bar < numBars; bar++)
				if (ok[foo][bar])
					if (vacateBar(bar)) {
						barFor[foo] = bar;
						fooFor[bar] = foo;
						fooTried[foo] = false;
						return true;
					}
			return false;
		} else
			return false;
	}
	
	private boolean vacateBar(int bar) {
		//if (!barTried[bar]) {
		//	barTried[bar] = true;
			return (fooFor[bar] == UNMATCHED || rematchFoo(fooFor[bar]));
		//} else
		//	return false;
	}
	
	// For the ensuring-certain-bars-are-used phase
	private boolean unmatchFoo(int foo) {
		return (barFor[foo] == UNMATCHED || fillBar(barFor[foo]));
	}
	
	private boolean fillBar(int bar) {
		if (!barRequired[bar]) {
			fooFor[bar] = UNMATCHED;
			return true;
		} else if (!barTried[bar]) {
			barTried[bar] = true;
			for (int foo = 0; foo < numFoos; foo++)
				if (ok[foo][bar])
					if (unmatchFoo(foo)) {
						fooFor[bar] = foo;
						barFor[foo] = bar;
						barTried[bar] = false;
						return true;
					}
			return false;
		} else
			return false;
	}
	
	public boolean augment() {
		fooTried = new boolean[numFoos];
		barTried = new boolean[numBars];
		
		/*augmentLoop: for (numMatched = 0; ; numMatched++) {
			fill(fooTried, false);
			//fill(barTried, false);
			for (int foo = 0; foo < numFoos; foo++)
				if (barFor[foo] == UNMATCHED && rematchFoo(foo))
					continue augmentLoop;
			break;
		}*/
		
		// Can we get away with doing this once outside the loop?
		fill(fooTried, false);
		for (int foo = 0; foo < numFoos; foo++) {
			if (barFor[foo] == UNMATCHED) {
				if (rematchFoo(foo))
					numMatched++;
			}
		}
		
		fill(barTried, false);
		for (int bar = 0; bar < numBars; bar++)
			if (fooFor[bar] == UNMATCHED && barRequired[bar]) {
				if (!fillBar(bar))
					return false;
			}
		return true;
	}
	
	private void evenFoo(int foo) {
		switch (fooGed[foo]) {
		case O:
			throw new AssertionError();
		case U:
			fooGed[foo] = GED.E;
			for (int bar = 0; bar < numBars; bar++)
				if (ok[foo][bar] && bar != barFor[foo])
					oddBar(bar);
		case E:
		}
	}
	
	private void oddFoo(int foo) {
		switch (fooGed[foo]) {
		case E:
			throw new AssertionError();
		case U:
			fooGed[foo] = GED.O;
			if (barFor[foo] != UNMATCHED)
				evenBar(barFor[foo]);
		case O:
		}
	}
	
	private void evenBar(int bar) {
		switch (barGed[bar]) {
		case O:
			throw new AssertionError();
		case U:
			barGed[bar] = GED.E;
			for (int foo = 0; foo < numFoos; foo++)
				if (ok[foo][bar] && foo != fooFor[bar])
					oddFoo(foo);
		case E:
		}
	}
	
	private void oddBar(int bar) {
		switch (barGed[bar]) {
		case E:
			throw new AssertionError();
		case U:
			barGed[bar] = GED.O;
			if (fooFor[bar] != UNMATCHED)
				evenFoo(fooFor[bar]);
		case O:
		}
	}
	
	public void decompose() {
		fill(fooGed, GED.U);
		fill(barGed, GED.U);
		for (int foo = 0; foo < numFoos; foo++)
			if (barFor[foo] == UNMATCHED)
				evenFoo(foo);
		for (int bar = 0; bar < numBars; bar++)
			if (fooFor[bar] == UNMATCHED)
				evenBar(bar);
	}
}
